import java.util.*;
import java.io.*;

/**
 * This class represents the dictionary generated by DicExtractor and handed on to Replacer for conversion
 * @author Johannes Dellert <jdellert@sfs.uni-tuebingen.de>
 * @version 0.1, November 2007
 */

public class ConversionDictionary
{
    //this structure stores all the information: POS Strings are hashed to hashmaps which hash orth Strings to their conversions
    private HashMap<String,HashMap<String,String>> dict;

    /**
     * create a new conversion dictionary to store masking information sorted by POS classes
     */
    public ConversionDictionary()
    {
        //empty dictionary for strip and x-replace mode
        dict = new HashMap<String,HashMap<String,String>>();
        //this part is needed if no POS classes are separated
        dict.put("", new HashMap<String,String>());
    }
    
    /**
     * add a new entry (= the information from one leaf)
     * @param pos - the new entry's POS class
     * @param orth - the new entry's original version
     * @param conv - the new entry's masked version
     */
    public void addEntry(String pos, String orth, String conv)
    {
        //default category: empty string
        if (pos == null) pos = "";
        //create new subdictionary for each POS
        if (dict.get(pos) == null)
        {
            dict.put(pos, new HashMap<String,String>());
        }
        //if within the corresponding POS subdictionary orth String is not yet present, map it to the conversion handed over
        if (dict.get(pos).get(orth) == null)
        {
            dict.get(pos).put(orth,conv);
        }
    }
    
    /**
     * retrieve the replace token for an original token
     * @param pos - the token's POS class
     * @param orth - the token
     * @return - the masked version of the token from the dictionary
     */
    public String getConv(String pos, String orth)
    {
        if (dict.get(pos) != null)
        {
            return dict.get(pos).get(orth);
        }
        //default category: empty string
        return dict.get("").get(orth);
    }
    
    /**
     * used by the morphology analyzer to  get more direct access to the dictionary
     * @return - the dictionary structure to be processed by the morphology analyzer
     */
    public HashMap<String,HashMap<String,String>> getMap()
    {
        return dict;
    }

    /**
     * simply print out list sorted according to POS tags: POS original replacement
     * @param outputFile - the file to write the dictionary to
     */
    public void writeToFile(File outputFile) throws IOException, NullPointerException
    {
        String presentLine = "";
        FileWriter out = new FileWriter(outputFile);
        for (String pos : dict.keySet())
        {
            for (String orth : dict.get(pos).keySet())
            {
                presentLine += pos + " " + orth + " " + dict.get(pos).get(orth) + "\n";
                out.write(presentLine);
                presentLine = "";
            }
        }
        out.close();
    }

    /**
     * read a dictionary from a file as generated by writeToFile()
     * @param keyFile - the file to write the dictionary to
     */   
    public static ConversionDictionary read(File keyFile)
    {
        System.out.print("Now loading dictionary...");
        ConversionDictionary newDict = new ConversionDictionary();
        try
        {
            Scanner  in = new Scanner(keyFile);
            Scanner lineScanner;
            String pos = "";
            String key = "";
            String value = "";
            while (in.hasNextLine())
            {
                lineScanner = new Scanner(in.nextLine());
                if (lineScanner.hasNext()) pos = lineScanner.next();
                if (lineScanner.hasNext()) value = lineScanner.next();
                if (lineScanner.hasNext()) key = lineScanner.next();    
                newDict.addEntry(pos,value,key);
            }
        }
        catch (IOException e)
        {
            System.out.println("Couldn't read key file. Aborting ...");
            System.exit(1);
        }
        System.out.println("done.");
        return newDict;
    }

    /**
     * read a dictionary in reverse from a file as generated by writeToFile()
     * dictionary is then in the correct format to be used for demasking
     * @param keyFile - the file to write the dictionary to
     */
    public static ConversionDictionary readReverse(File keyFile)
    {
        System.out.print("Now loading dictionary...");
        ConversionDictionary newDict = new ConversionDictionary();
        try
        {
            Scanner  in = new Scanner(keyFile);
            Scanner lineScanner;
            String pos = "";
            String key = "";
            String value = "";
            while (in.hasNextLine())
            {
                lineScanner = new Scanner(in.nextLine());
                if (lineScanner.hasNext()) pos = lineScanner.next();
                if (lineScanner.hasNext()) value = lineScanner.next();
                if (lineScanner.hasNext()) key = lineScanner.next();    
                newDict.addEntry(pos,key,value);
            }
        }
        catch (IOException e)
        {
            System.out.println("Couldn't read key file. Aborting ...");
            System.exit(1);
        }
        System.out.println("done.");
        return newDict;
    }
} 
