import java.io.*;
import java.util.*;
import javax.xml.parsers.*;

/**
 * A console application to mask a corpus XML file using a wide range of different methods
 * @author Johannes Dellert <jdellert@sfs.uni-tuebingen.de>
 * @version 0.1, November 2007
 */

public class CorpusMasker
{
    /**
     * This program serves to mask a corpus XML file using a wide range of different methods
     * Usage:   CorpusMasker (options) [XML-File] [output file]
     *  With the following options:
     *          --userDefPaths           to let the user specify where to find the information in the trees
     *
     *          --extendDic              to specify a dictionary that is going to be used for masking
     *          
     *          --ccf                    to specify a character class file that will determine how replacement patterns are generated
     *
     *          --strip          to completely delete the text from the trees
     *          --x-replace      to replace the text with x characters
     *          --buildDic       to build a dictionary of random replace strings
     *          --buildDic-aff   to build a dictionary of random replace strings with preserved morphological information [default]
     *
     *          --preservePOS=XX,XXX,X   to exclude selected POS classes from being masked
     *          --noMorph=XX,XXX,X       to exclude selected POS classes from affix extraction
     *
     *          --affixDefOccPerWord=[number between 0 and 1]    to determine how greedy affix extraction will be
     *          --affixDefMinOccurs=[integer number]     to determine how often an affix has to occur for being accepted
     *
     *          --verbose        for verbose output of the dictionary building process
     */
    public static void main(String[] args)
    {
        //we always need two files: input file (xmlFile) and output file
        File xmlFile = null;
        File outputFile = null;
        //sometimes we need an external dict file
        String dictFileName = null;
        //where to get the character class information from; defaults to German
        String ccfFileName = "lang/german.ccf";
        //there are four modes of anonymizing
        int mode = 4;
    	//two numeric parameters for affix extraction
    	double occPerWord = 0.02;
    	int minOccurs = 1;
    	//two sets of strings define which POS classes are to be processed
    	Set<String> preservedPOS = new HashSet<String>();
    	Set<String> noMorphPOS = new HashSet<String>();
        //these strings specify where in the input document the decisive information is to be found
        String orthPath = "../orth";
        String posPath = "../pos";
        //are locations defined by the user?
        boolean userDefPaths = false;
        //verbose output disabled by default
        boolean verboseOutput = false;
        //is there an existing dict file to be used?
        boolean extendDic = false;
        //check for correct argument structure and read out arguments
        if (args.length < 2)
        {
            printUsage();
            System.exit(1);
        }
        if (args.length == 2)
        {
            if (args[0].startsWith("--"))
            {
                printUsage();
                System.exit(1);
            }
            xmlFile = new File(args[0]);
            outputFile = new File(args[1]);
        }
        //last two will be interpreted as filenames, the rest are options
        if (args.length >= 3)
        {
            for(int i = 0; i < args.length - 2; i++)
            {
                String presentArgument = args[i];
                // enable user defined tags to ensure compatibility with many corpora
                if (presentArgument.equals("--userDefPaths")) userDefPaths = true;
                // enable user to define an existing conversion dictionary to be extended
                if (presentArgument.equals("--extendDic")) extendDic = true;
                // the strip mode: simply delete chars within orth elements
                else if (presentArgument.equals("--strip")) mode = 1;
                // the x-replace mode: replace all chars within orth elements with X or x (depending on case)
                else if (presentArgument.equals("--x-replace")) mode = 2;
                //in this mode, a dictionary of random replacements is built (to make replacements countable etc.)
                else if (presentArgument.equals("--buildDic")) mode = 3;
                //in this mode, morphological information (affixes) are tried to be kept in the dictionary
                else if (presentArgument.equals("--buildDic-aff")) mode = 4;
                //give user possibility to specify another ccf file to define character classes
                else if (presentArgument.equals("ccf"))
                {
                    System.out.print("Please specify ccf File ");
                    Scanner fileNameScanner = new Scanner(System.in);
                    ccfFileName = fileNameScanner.next(); 
                }
                //read out POS classes to be preserved
                else if (presentArgument.startsWith("--preservePOS="))
                {
                    Scanner listScanner = new Scanner(presentArgument.substring(14)).useDelimiter(",");
                    while (listScanner.hasNext())
                    {
                        preservedPOS.add(listScanner.next());
                    }
                    System.out.println("The following POS classes will be excluded from being masked:");
                    for (String pos : preservedPOS)
                    {
                        System.out.print(pos + " ");
                    }
                    System.out.println("");
                }
                //read out POS classes to be excluded from morphological analysis
                else if (presentArgument.startsWith("--noMorph="))
                {
                    Scanner listScanner = new Scanner(presentArgument.substring(10)).useDelimiter(",");
                    while (listScanner.hasNext())
                    {
                        noMorphPOS.add(listScanner.next());
                    }
                    System.out.println("The following POS classes will be excluded from affix extraction:");
                    for (String pos : noMorphPOS)
                    {
                        System.out.print(pos + " ");
                    }
                    System.out.println("");
                }
                //read out ratio of affix occurrence and total word number in each class as a threshold value of affix recognition
                else if (presentArgument.startsWith("--affixDefOccPerWord="))
                {
                    try
                    {
                        double temp = Double.parseDouble(presentArgument.substring(22));
                        if (temp < 0 || temp > 1)
                        {
                            System.out.println("Invalid parameter: " + presentArgument.substring(22) + ". Value must be between 0 and 1.");
                            System.exit(1);
                        }
                        else
                        {
                            occPerWord = temp;
                        }
                    }
                    catch(Exception e)
                    {
                        System.out.println("Invalid parameter: " + presentArgument.substring(22) + ". Double value expected.");
                        System.exit(1);
                    }
                }
                //read out minimum number of occurrences of each affix to be accepted
                else if (presentArgument.startsWith("--affixDefMinOccurs="))
                {
                    try
                    {
                        int temp = Integer.parseInt(presentArgument.substring(20));
                        minOccurs = temp;
                    }
                    catch(Exception e)
                    {
                        System.out.println("Invalid parameter: " + presentArgument.substring(20) + ". Integer value expected.");
                        System.exit(1);
                    }
                }
                // define whether the DictExtractor will deliver verboseOutput
                else if (presentArgument.equals("--x-replace")) verboseOutput = true;
                else
                {
                    System.out.println("Invalid argument: " + presentArgument + ". Proceeding with default option ...");
                }
            }
            xmlFile = new File(args[args.length - 2]);
            outputFile = new File(args[args.length - 1]);          
        }
        //let the user define the tags if desired
        if (userDefPaths)
        {
            System.out.print("Please specify the XPath to the elements containing the words: ");
            Scanner tagNameScanner = new Scanner(System.in);
            orthPath = tagNameScanner.next(); 
            System.out.print("Please specify the XPath to the elements containing the POS tags: ");
            tagNameScanner = new Scanner(System.in);
            posPath = tagNameScanner.next(); 
        }
        //read in an external ConvDict to be extended
        if (extendDic)
        {
            System.out.print("Please give the file name of the dictionary file to be extended: ");
            Scanner fileNameScanner = new Scanner(System.in);
            dictFileName = fileNameScanner.next();
        }
        //this first SAX Parser will extract the words and POS tags and create the dictionary
        DicExtractor extractor = new DicExtractor(orthPath, posPath, preservedPOS, verboseOutput, ccfFileName);
        //the second SAX Parser is responsible for actually processing the document
        Replacer replacer = new Replacer(outputFile, mode, orthPath, posPath, null);
        boolean loadSuccess = xmlFile.canRead();
        if (loadSuccess)
        {
            try
            {
                SAXParserFactory factory = SAXParserFactory.newInstance();
                SAXParser saxParser = factory.newSAXParser();
                //dictionary building process only needed for modes 3 and 4
                if (mode >= 3)
                {
                    //this line is for loading an external ConvDict
                    if (extendDic)
                    {
                        extractor.setConvDict(ConversionDictionary.read(new File(dictFileName)));
                    }
                    //extract the dictionary from the input file
                    saxParser.parse(xmlFile, extractor);
                    //reestablish morphology if in mode 4; parameters will determine degree of masking
                    if (mode == 4) extractor.retainMorphology(occPerWord,minOccurs,noMorphPOS);
                    //hand on the ready dictionary to the replacer
                    replacer.setConvDict(extractor.getConvDict());
                    //give user possibility to store dictionary on hard disk; needed for later deanonymizing
                    System.out.print("Please specify name for new dictionary file: ");
                    Scanner fileNameScanner = new Scanner(System.in);
                    extractor.getConvDict().writeToFile(new File(fileNameScanner.next())); 
                }
                //now let the second parser replace the orth elements by means of the dictionary
                saxParser.parse(xmlFile, replacer);
            }
            catch (Throwable t)
            {
                t.printStackTrace();
            }
        }
        else
        {
            System.out.println("File " + xmlFile + " does not exist!");
        }
    }

    private static void printUsage()
    {
        System.out.println("Usage:\t CorpusMasker (options) [XML-File] [output file]");
        System.out.println("\tWith the following options:\n");
        System.out.println("\t\t--userDefPaths\t\t to let the user specify where to find the information in the trees\n");
        System.out.println("\t\t--extendDic\t\t to specify a dictionary that is going to be used for masking\n");
        System.out.println("\t\t--ccf\t\t to specify a character class file that will determine how replacement patterns are generated\n");       
        System.out.println("\t\t--strip\t\t to completely delete the text from the trees");
        System.out.println("\t\t--x-replace\t to replace the text with x characters");
        System.out.println("\t\t--buildDic\t to build a dictionary of random replace strings");
        System.out.println("\t\t--buildDic-aff\t to build a dictionary of random replace strings with preserved morphological information [default]\n");
    	System.out.println("\t\t--preservePOS=XX,XXX,X\t to exclude selected POS classes from being masked");
    	System.out.println("\t\t--noMorph=XX,XXX,X\t to exclude selected POS classes from affix extraction\n");
    	System.out.println("\t\t--affixDefOccPerWord=[number between 0 and 1]\t to determine how greedy affix extraction will be");
    	System.out.println("\t\t--affixDefMinOccurs=[integer number]\t to determine how often an affix has to occur for being accepted\n");
    	System.out.println("\t\t--verbose\t for verbose output of the dictionary building process\n");
    }
} 
