import java.io.File;
import java.io.FileNotFoundException;
import java.util.*;

import org.xml.sax.*;
import org.xml.sax.helpers.*;

/**
 * This SAX parser reads through the first part of an XML corpu doument
 * reading all the orth elements together with their POS tags and storing
 * the replacements it generates in a ConversionDictionary
 * until it achieves a sample size of 1000 masking dictionary entries
 * @author Johannes Dellert <jdellert@sfs.uni-tuebingen.de>
 * @version 0.1, November 2007
 */


public class DicSampleExtractor extends DefaultHandler 
{
    //necessary to know where we are during sequential parsing
    boolean isOrth;
    boolean isPos;
    String presentOrth;
    String presentPos;

    //necessary to find elements to extract
    String orthPath;
    String posPath;
    XPathTracker tracker;

    ConversionDictionary dict;

    static int MAX_TRIES = 100;
    
    //specify letter classes to be interchanged
    HashMap<String,ArrayList<String>> charClasses;

    //necessary for correct processing of POS classes
    Set<String> presPOS;

    //set to true in order to see debugging information
    boolean verbose;
    
    int entries;

    /**
     * A DictSampleExtractor needs specification of the content to be masked and a definition of untouchable content as well as a file defining the char classes for replacement.
     * @param theOrthPath - the XPath expression to define masked content
     * @param thePosPath - the XPath expression to define classification of content for dictionary retrieval
     * @param preservedPOS - a set of POS classes whose members are exempt from being masked
     * @param verboseOutput - true for very verbose output of the building process
     * @param ccfFileName - the name of a file in the format exemplified by lang/german.ccf 
     */
    public DicSampleExtractor(String theOrthPath, String thePosPath, Set<String> preservedPOS, boolean verboseOutput, String ccfFileName)
    {
        super();
        isOrth = false;
        isPos = false;
        presentOrth = "";
        presentPos = "";
        orthPath = theOrthPath;
        posPath = thePosPath;
        dict = new ConversionDictionary();
        presPOS = preservedPOS;
        verbose = verboseOutput;
        tracker = new XPathTracker();
        entries = 0;
        charClasses = loadCharClasses(ccfFileName); 
    }

    /**
     * get the masking dictionary generated from the document while traversing it
     * @return the ConversionDictionary generated while traversing the document and generating replacement patterns
     */
    public ConversionDictionary getConvDict()
    {
        return dict;
    }
    
    /**
     * specify the char classes for replace pattern generation by giving a file in the format exemplified by lang/german.ccf
     * @param fileName - a filename leading to a file in the format exemplified by lang/german.ccf 
     * @return - a map with character class information
     */
    public static HashMap<String,ArrayList<String>> loadCharClasses(String fileName)
    {
        HashMap<String,ArrayList<String>> cClasses = new HashMap<String,ArrayList<String>>();
        File ccFile = new File(fileName);
        try
        {
            Scanner s = new Scanner(ccFile);
            Scanner lineScanner;
            while (s.hasNextLine())
            {
                String line = s.nextLine();
                lineScanner = new Scanner(line);
                if (line.startsWith("PUNCT"))
                {
                    lineScanner.next();
                    ArrayList<String> punct = new ArrayList<String>();
                    while (lineScanner.hasNext())
                    {
                        punct.add(lineScanner.next());
                    }
                    cClasses.put("punct", punct);
                }
                else if (line.startsWith("DIGITS"))
                {
                    lineScanner.next();
                    ArrayList<String> digits = new ArrayList<String>();
                    while (lineScanner.hasNext())
                    {
                        digits.add(lineScanner.next());
                    }
                    cClasses.put("digits", digits);
                }
                else
                {
                    ArrayList<String> charClass = new ArrayList<String>();
                    while (lineScanner.hasNext())
                    {
                        String nextChar = lineScanner.next();
                        if (nextChar.startsWith("(") && nextChar.endsWith(")"))
                        {
                            cClasses.put(nextChar.substring(1,nextChar.length() - 1), charClass);
                        }
                        else
                        {
                            charClass.add(nextChar);
                            cClasses.put(nextChar, charClass);
                        }
                    }
                }
            }
            return cClasses;
        }
        catch (FileNotFoundException e)
        {
            System.out.println("Error: Unable to load Character Class File " + fileName);
            System.exit(1);
        }   
        return null;
    }

    public void startDocument()
    {
        System.out.print("Scanning through input file and building the dictionary... ");
    }

    public void endDocument()
    {
        System.out.println("done");
    }

    public void startElement(String namespaceURI, String sName, String qName, Attributes attrs) throws SAXException
    {
        if (posPath.equals(""))
        {
            presentPos = "";
        }
        //keep the XPath and the Orth-POS flags up to date
        tracker.goToElement(qName);
        if (tracker.atPath(orthPath)) isOrth = true;
        if (tracker.atPath(posPath)) isPos = true;
        //extract Orth-POS pair from attributes if necessary
        if (attrs.getLength() >= 1)
        {
            for (int i = 0; i < attrs.getLength(); i++)
            {
                tracker.goToAttribute(attrs.getQName(i));
                if (tracker.atPath(orthPath)) presentOrth = new String(attrs.getValue(i)).toLowerCase();
                if (tracker.atPath(posPath)) presentPos = new String(attrs.getValue(i));
                updateDictionary();
                tracker.levelUp();
            }
        }
    }

    public void endElement(String namespaceURI, String sName, String qName) throws SAXException
    {
        //keep the XPath and the Orth-POS flags up to date
        if (tracker.atPath(orthPath)) 
        {
            isOrth = false;
            updateDictionary();
        }
        if (tracker.atPath(posPath)) 
        {
            isPos = false;
            updateDictionary();
        }
        tracker.levelUp();
    }

    public void characters(char[] ch, int start, int length)
    {
        //simply store content of each orth element and check whether a complete Orth-POS pair is full
        if (isOrth)
        {
            if (presentOrth == null)
            {
                presentOrth = new String(ch, start, length);
            }
            else
            {
                presentOrth += new String(ch, start, length);
            }
            //part of uppercase handling: dictionary only contains lowercase entries
            presentOrth = presentOrth.toLowerCase();
        }
        //also store content of any pos element and check whether a complete Orth-POS pair is full
        if (isPos)
        {
            if (presentPos == null)
            {
                presentPos = new String(ch, start, length);
            }
            else
            {
                presentPos += new String(ch, start, length);
            }
        }
    }

    //this method handles the conversion into replacement patterns and the insertion into the directory
    private void updateDictionary() throws SAXException
    {
        //if we have a complete Orth-POS pair
        if ((presentPos != null) && (presentOrth != null))
        {
            if (verbose) System.out.println("Now processing '" + presentOrth + "'");
            String converted;
            if (dict.getMap().get(presentPos) != null)
            {   
                //don't generate new patterns if original is already in dict
                if (dict.getMap().get(presentPos).get(presentOrth) == null)
                {
                    //for preserved word classes, simply define original as conversion
                    if (presPOS.contains(presentPos))
                    {
                        //output for debugging purposes: decomment to get verbose output on replacement patterns
                        if (verbose) System.out.println(presentOrth + " : " + presentPos + " preserved");
                        dict.addEntry(presentPos, presentOrth, presentOrth);
                        entries++;
                        //no further processing in this case
                        presentPos = null;
                        presentOrth = null;
                        return;
                    }
                    //randomly convert the string as defined in the convert methods
                    converted = convert(presentOrth);
                    //don't use patterns that do already exist, but make at maximum MAX_TRIES tries to find another pattern
                    int tries = 0;
    
                    while (dict.getMap().get(presentPos).values().contains(converted) && tries < MAX_TRIES)
                    {
                            if (verbose) System.out.println("Found '" + converted + "' in dict for POS " + presentPos + ".");
        
                            //reconvert and compare
                            converted = convert(converted);

                            tries++;

                            //output for debugging purposes: protocols the attempts to find a replacement pattern
                            if (verbose) System.out.println("Attempt #" + tries + ". Now trying: " + converted);
                    }
    
                    //output for debugging purposes: decomment to get verbose output on replacement patterns
                    if (verbose) System.out.println(presentOrth + " : " + presentPos + " ---> " + converted);
                    dict.addEntry(presentPos, presentOrth, converted);
                    entries++;
                    presentPos = null;
                    presentOrth = null;
                }
                else
                {
                    if (verbose) System.out.println("Entry for '" + presentOrth + "' already present: Skipping");
                    presentPos = null;
                    presentOrth = null;
                    return;
                }
            }
            else
            {
                //randomly convert the string as defined in the convert methods
                converted = convert(presentOrth);     
                //output for debugging purposes: decomment to get verbose output on replacement patterns
                if (verbose) System.out.println(presentOrth + " : " + presentPos + " ---> " + converted);
                dict.addEntry(presentPos, presentOrth, converted);
            }
            presentPos = null;
            presentOrth = null;
        }
        if (entries > 1000)
        {
            throw new SAXException("Sample complete!");
        }
    }

    /**
     * find a replacement pattern for a token 
     * @param toConvert - the string to convert
     * @return - the converted String
     */
    public String convert (String toConvert)
    {
        String output = "";
        for (int i = 0; i < toConvert.length(); i++)
        {
            char convertedChar = convertChar(toConvert.charAt(i));
            //we generally do not want chars to stay the same
            while (convertedChar == toConvert.charAt(i))
            {
                //punctuation, digits and one-letter tokens may stay the same
		if (isPunctuation(convertedChar) || isNumber(convertedChar) || toConvert.length() == 1) 
                {
                    if (verbose) System.out.println("\tChar stayed the same. Character: " + convertedChar + "Skipped.");   
                    break;    
                }
                if (verbose) System.out.println("\tChar stayed the same. Problem character: " + convertedChar);   
                convertedChar = convertChar(convertedChar);
            }
            output += convertedChar;
        }
        return output;
    }

    //helper method to define punctuation
    private boolean isPunctuation(char testChar)
    {
        return charClasses.get("punct").contains(Character.toString(testChar));
    }

    //helper method to define digits
    private boolean isNumber(char testChar)
    {
        return charClasses.get("digits").contains(Character.toString(testChar));
    }

    /**
     * convert a character using the char class definition
     * @param toConvert - the char to convert
     * @return - the converted char
     */
    public char convertChar(char toConvert)
    {
        ArrayList<String> replacementCandidates = charClasses.get(Character.toString(toConvert));
        if (replacementCandidates == null) return toConvert;
        int randomReplacement = (int)(Math.random() * replacementCandidates.size());
        return replacementCandidates.get(randomReplacement).charAt(0);
    }

    /**
     * revise the whole dictionary to detect affixes for each POS class
     * and reinsert them to the replacement forms (morphological information will largely be preserved)
     * parameters serve to determine how permissive the algorithm evaluating the possible affixes will be
     * @param occPerWord - define how often an affix has to occur in its word class to be retained 
     * @param minOccurs - define the minimum number of times an affix has to occur to be retained
     * @param noMorphPOS - define closed POS classes where no morphology should be inferred
     */
    public void retainMorphology(double occPerWord, int minOccurs, Set<String> noMorphPOS)
    {
        //go through each POS subdictionary and operate on the data in it
        for (String pos : dict.getMap().keySet())
        {
            //don't do any processing if POS is in list of classes not to be analyzed
            if (noMorphPOS.contains(pos))
            {
                System.out.println("No affix extraction for " + pos + ".");
                continue;
            }
            System.out.print("Scanning " + pos + " for affixes... ");
            //first, store all the mappings for this POS
            HashMap<String,String> posMappings = dict.getMap().get(pos);

            //here, the typical affixes of this POS are going to be stored together with their number of occurrences
            HashMap<String,Integer> prefix = new HashMap<String,Integer>();
            HashMap<String,Integer> suffix = new HashMap<String,Integer>();

            //simply store all beginning substrings of the words with this POS in a huge hash map and count their occurrences
            for (String orthForm : posMappings.keySet())
            {
                for (int i = 0; i < orthForm.length(); i++)
                {
                    String presentPrefix = orthForm.substring(0,i + 1);
                    if (prefix.get(presentPrefix) != null)
                    {
                        prefix.put(presentPrefix, (prefix.get(presentPrefix) + 1));
                    }
                    else
                    {
                        prefix.put(presentPrefix, 1);
                    } 
                }
            }

            //simply store all ending substrings of the words with this POS in a huge hash map and count their occurrences
            for (String orthForm : posMappings.keySet())
            {
                for (int i = orthForm.length(); i > 0; i--)
                {
                    String presentSuffix= orthForm.substring(i - 1);
                    if (suffix.get(presentSuffix) != null)
                    {
                        suffix.put(presentSuffix, (suffix.get(presentSuffix) + 1));
                    }
                    else
                    {
                        suffix.put(presentSuffix, 1);
                    } 
                }
            }

            //now discard all prefixes that don't exceed the specified threshold value of 
            //occurrences per word with this POS and do not appear the specified minimum number of times
            Object[] prefixes = prefix.keySet().toArray();
            String currentPrefix = "";
            for (int i = 0; i < prefixes.length; i++)
            {
                currentPrefix = (String) prefixes[i];
                if ((prefix.get(currentPrefix) < posMappings.size() * occPerWord) || (prefix.get(currentPrefix) < minOccurs))
                {
                    prefix.remove(currentPrefix);
                }
            }

            //now discard all suffixes that don't exceed the specified threshold value of 
            //occurrences per word with this POS and appear the specified minimum number of times
            Object[] suffixes = suffix.keySet().toArray();
            String currentSuffix = "";
            for (int i = 0; i < suffixes.length; i ++)
            {
                currentSuffix = (String) suffixes[i];
                if ((suffix.get(currentSuffix) < posMappings.size() * occPerWord) || (suffix.get(currentSuffix) < minOccurs))
                {
                    suffix.remove(currentSuffix);
                }
            }

            System.out.println("found " + prefix.size() + " prefixes, " + suffix.size() + " suffixes.");

            //the following part discommented will print out the found affixes for each POS class
            if (verbose)
            {
                System.out.println("Prefixes for " + pos + ":");
                for (String presentPrefix : prefix.keySet())
                {
                    System.out.println("\t" + presentPrefix + ": " + prefix.get(presentPrefix) + " instances");
                }
                System.out.println("\nSuffixes for " + pos + ":");
                for (String presentSuffix : suffix.keySet())
                {
                    System.out.println("\t" + presentSuffix + ": " + suffix.get(presentSuffix) + " instances");
                }
                System.out.println("");
            }
            System.out.print("Now revising the dictionary to retain these affixes... ");
            //go through all the words of this POS tag again
            for (String orthForm : posMappings.keySet())
            {
                //for every found prefix, check whether word begins with it ...
                for (String presentPrefix : prefix.keySet())
                {
                    if (orthForm.startsWith(presentPrefix))
                    {
                        //... and retain the prefix in the replacement form
                        String newConversion = new String(presentPrefix);
                        newConversion += posMappings.get(orthForm).substring(presentPrefix.length());
                        //if ambiguities arise, don't apply morphology in order to disambiguate
                        if (!dict.getMap().get(pos).values().contains(newConversion))
                        {
                            posMappings.put(orthForm, newConversion);
                        }
                    }
                }
                //for every found suffix, check whether word begins with it ...
                for (String presentSuffix : suffix.keySet())
                {
                    if (orthForm.endsWith(presentSuffix))
                    {
                        //... and retain the suffix in the replacement form
                        String newConversion = posMappings.get(orthForm).substring(0, posMappings.get(orthForm).length() - presentSuffix.length());
                        newConversion += presentSuffix ;
                        //if ambiguities arise, don't apply morphology in order to disambiguate
                        if (!dict.getMap().get(pos).values().contains(newConversion))
                        {
                            posMappings.put(orthForm, newConversion);
                        }
                    }
                }
            }
            System.out.println("done.\n");
        }
    }
}
