import org.xml.sax.*;
import org.xml.sax.helpers.*;

/**
 * This SAX parser processes the beginning of an XML corpus file and generates a copy
 * replacing in mode 3 or 4 all orth elements with the replacements specified in a ConversionDictionary,
 * otherwise stripping the elements or replacing them with strings of Xx chars
 * changed parts are marked with HTML font color markup during the process
 * @author Johannes Dellert <jdellert@sfs.uni-tuebingen.de>
 * @version 0.1, November 2007
 */

public class ReplaceMarker extends DefaultHandler 
{
    int mode;
    
    //necessary to know where we are during sequential parsing
    boolean isOrth;
    boolean isPos;
    String presentOrth;
    String presentPos;
    XPathTracker tracker;

    //necessary to find elements to replace
    String orthPath;
    String posPath;


    //the dictionary delivered by the DicExtractor
    ConversionDictionary dict;

    //variables to handle output
    String origBufferedOutput;
    String bufferedOutput;
    StringBuffer in;
    StringBuffer out;
    
    //communication with the graphical user interface
    Drop logDrop;

    //after reading an orth element without knowing its pos, output will be stored in buffered output until replacement can be calculated
    boolean betweenElements;

    /**
     * A Replacer needs two String buffers for original and preview data, a mode, and specification of the content to mask.
     * @param input - a String buffer for constructing the HTML representation of the input data
     * @param output - a String buffer for constructing the HTML representation of the preview data
     * @param aMode - an integer between 0 and 4 for different modes (for internal use)
     * @param theOrthPath - the XPath expression to define masked content
     * @param thePosPath - the XPath expression to define classification of content for dictionary retrieval
     * @param log - a Drop object to handle logging messages in the GUI - null for console output
     */
    public ReplaceMarker(StringBuffer input, StringBuffer output, int aMode, String theOrthPath, String thePosPath, Drop log)
    {
        super();
        mode = aMode;

        isOrth = false;
        isPos = false;
        orthPath = theOrthPath;
        posPath = thePosPath;
        presentOrth = null;
        presentPos = null;
        dict = new ConversionDictionary();
        origBufferedOutput = "";
        bufferedOutput = "";
        betweenElements = false;
        tracker = new XPathTracker();
        in = input;
        out = output;
        logDrop = log;
    }

    /**
     * specify a ConversionDictionary to be used for masking
     * @param theDict - a ConversionDictionary defining how to mask the different tokens
     */
    public void setConvDict(ConversionDictionary theDict)
    {
        dict =  theDict;
    }

    public void startDocument()
    {
        in.append("<?xml version='1.0' encoding='utf-8'?>");
        out.append("<?xml version='1.0' encoding='utf-8'?>");
    }

    public void endDocument()
    {
    }

    public void startElement(String namespaceURI, String sName, String qName, Attributes attrs) throws SAXException
    {
            if (posPath.equals(""))
            {
                presentPos = "";
            }
            //update XPath and check whether element contains any necessary information
            tracker.goToElement(qName);
            if (tracker.atPath(orthPath)) isOrth = true;
            if (tracker.atPath(posPath)) isPos = true;
            
            origOutput("<" + qName);
            output("<" + qName);
            //attribute handling
            if (attrs.getLength() >= 1)
            {
                for (int i = 0; i < attrs.getLength(); i++)
                {
                    //what if info in attribute, i.e. an XPath containing @ sign?
                    //adjust present XPath
                    tracker.goToAttribute(attrs.getQName(i));
                    //check whether we are at the position for an orth tag 
                    if (tracker.atPath(orthPath)) 
                    {
                        presentOrth = attrs.getValue(i);
                        //if corresponding pos has not yet been scanned ...
                        if (presentPos == null) 
                        {
                            //... redirect output to buffer until pos found
                            betweenElements = true;
                        }
                        //if corresponding pos has already been scanned ...
                        else
                        {
                            //... simply output the attribute
                            betweenElements = false;
                            origOutput(" " + attrs.getQName(i) + "='" + presentOrth +"'");
                            output(" " + attrs.getQName(i) + "='" + lookUp(presentOrth) +"'");
                        }
                    }
                    else 
                    {
                        //check whether we are at the position for an orth tag 
                        if (tracker.atPath(posPath))
                        {
                            presentPos = new String(attrs.getValue(i));
                            //if corresponding orth element has already been scanned ...
                            if (betweenElements)
                            {
                                //store pos output in buffer
                                origOutput(" " + attrs.getQName(i) + "='" + attrs.getValue(i)+"'");
                                output(" " + attrs.getQName(i) + "='" + attrs.getValue(i)+"'");
                                //print orth attribute and empty buffer to file
                                betweenElements = false;
                                origOutput(" " + attrs.getQName(i) + "='" + presentOrth +"'");
                                output(" " + attrs.getQName(i) + "='" + lookUp(presentOrth) +"'");
                            }
                            //if corresponding orth element has not yet been scanned ...
                            else
                            {
                                //write info to file and store pos in presentPos
                                origOutput(" " + attrs.getQName(i) + "='" + attrs.getValue(i)+"'");
                                output(" " + attrs.getQName(i) + "='" + attrs.getValue(i)+"'");
                            }
                        }
                        //all other attributes can simply be written into the file
                        else
                        {
                            origOutput(" " + attrs.getQName(i) + "='" + attrs.getValue(i)+"'");
                            output(" " + attrs.getQName(i) + "='" + attrs.getValue(i)+"'");
                        }
                    }
                    tracker.levelUp();
                }
            }
            //don't copy orth tags, but reset presentOrth when encountering them
            if (tracker.atPath(orthPath)) 
            {
                isOrth = true;
            }
            //don't copy pos tags, but store their attributes
            else if (tracker.atPath(posPath)) 
            {
                isPos = true;
            }
            origOutput(">");
            output(">");
    }

    public void endElement(String namespaceURI, String sName, String qName) throws SAXException
    {
        //if it is an orth element
        if (tracker.atPath(orthPath)) 
        {
            isOrth = false;
            if (presentPos == null)
            {
                betweenElements = true;
            }
            else
            {
                betweenElements = false;
                //mark the replacement in red in the original and in green in the preview
                origOutput("[[font color=\"red\"]]" + presentOrth + "[[/font]]");
                output("[[font color=\"green\"]]" + lookUp(presentOrth) + "[[/font]]");
            }
        }
        //if it is a pos element...
        if (tracker.atPath(posPath)) 
        {
            isPos = false;
            //if no POS could be found, do not print "null"
            if (presentPos == null) presentPos = "";
            //... completing one orth-pos pair, do the conversion and output the buffered string
            if (betweenElements)
            {
                //mark the detected POS category in blue in the original and in the preview
                origOutput("[[font color=\"blue\"]]" + presentPos + "[[/font]]");
                output("[[font color=\"blue\"]]" + presentPos + "[[/font]]");
                betweenElements = false;
                //mark the replacement in red in the original and in green in the preview
                origOutput("[[font color=\"red\"]]" + presentOrth + "[[/font]]");
                output("[[font color=\"green\"]]" + lookUp(presentOrth) + "[[/font]]");
            }
            //... starting one orth-pos pair, it can be directly put out because it won't change
            else
            {
                origOutput(presentPos);
                output(presentPos);
            }
        }
        //adjust the XPath position
        tracker.levelUp();
        //simply put out the ending tag
        origOutput("</" + qName + ">");
        output("</" + qName + ">");
  
    }

    //chars are copied into output file unless they belong to the desired orth or pos element
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        if (isOrth)
        {
            if (presentOrth == null)
            {
                presentOrth = new String(ch, start, length);
            }
            else
            {
                presentOrth += new String(ch, start, length);
            }
        }
        else if (isPos)
        {
            if (presentPos == null)
            {
                presentPos = new String(ch, start, length);
            }
            else
            {
                presentPos += new String(ch, start, length);
            }
        }
        else
        {
            origOutput(new String(ch, start, length));
            output(new String(ch, start, length));
        }
    }

    //output dependent on betweenElements flag; if output cannot be completed during linear parsing, output will be detained in buffer
    private void output(String toPrint) throws SAXException
    {
        if (!betweenElements)
        {
            //replacement necessary to handle &amp; entity
            out.append(toPrint.replace("&","&amp;"));
            out.append(bufferedOutput.replace("&","&amp;"));
            bufferedOutput = "";
        }
        else
        {
            bufferedOutput += toPrint;
        }
        if (out.length() > 20000) throw new SAXException("Complete");
    }
    
    private void origOutput(String toPrint)
    {
        if (!betweenElements)
        {
            //replacement necessary to handle &amp; entity
            in.append(toPrint.replace("&","&amp;"));
            in.append(origBufferedOutput.replace("&","&amp;"));
            origBufferedOutput = "";
        }
        else
        {
            origBufferedOutput += toPrint;
        }
    }

    //helper method to convert into replacements
    //ATTENTION: each call will empty presentPos and presentOrth to enable search for next POS-Orth pair
    private String lookUp(String theWord)
    {
        String theReplacement = "";
        //default POS category: any
        if (presentPos == null) presentPos = "any";
        //use dictionary to retrieve replace form
        if (mode >= 3)
        {
            //part of uppercase handler: store the uppercase pattern of the present form
            String casingPattern = xConvert(theWord);
            String newOrth = dict.getConv(presentPos, theWord.toLowerCase());
            if (newOrth == null)
            {
                log("Warning: Couldn't resolve: " + presentPos + " " + theWord + ". Please make sure the dictionary is complete.");
            }
            else
            {
                //part of uppercase handler: restore uppercase in retrieved form and return result
                theReplacement = applyCasingPattern(newOrth, casingPattern);
            }
        }
        //use local method to convert string to x chars
        if (mode == 2)
        {
            theReplacement = xConvert(theWord);
        }
        //simply leave the orth elements empty for mode 1

        //reset the two strings so that we can again test whether we have all the information we need
        presentPos = null;
        presentOrth = null;

        return theReplacement;
    }

    //this method converts input strings to strings of x and X (depending on casing)
    private String xConvert(String inputString)
    {
        String output = "";
        char presentChar;
        for (int i = 0; i < inputString.length(); i++)
        {
            presentChar = inputString.charAt(i);
            //for ASCII chars between 97 and 122 (lowercase letters a-z)
            if ((((int) presentChar >= 97)&&((int) presentChar <= 122))||(presentChar == 'ä')||(presentChar == 'ö')||(presentChar == 'ü')||(presentChar == 'ß'))
            {
                output += 'x';
            } 
            //for ASCII chars between 65 and 90 (uppercase letters A-Z)
            else if ((((int) presentChar >= 65)&&((int) presentChar <= 90))||(presentChar == 'Ä')||(presentChar == 'Ö')||(presentChar == 'Ü'))
            {
                output += 'X';
            }
            else  
            {
                output += presentChar;
            }
        }
        return output;
    }

    //this method changes the casing of an input string according to the casing pattern
    private String applyCasingPattern(String theWord, String thePattern)
    {
        String newWord = "";
        char presentChar;
        char presentCase;
        for (int i = 0; i < theWord.length() && i < thePattern.length(); i++)
        {
            presentChar = theWord.charAt(i);
            presentCase = thePattern.charAt(i);
            //if an uppercase letter is needed
            if (presentCase == 'X')
            {
                //convert lowercase to Uppercase
                if (((int) presentChar >= 97)&&((int) presentChar <= 122))
                {
                    newWord += (char) ((int) presentChar - 32);
                } 
                else if (presentChar == 'ä') newWord += 'Ä';
                else if (presentChar == 'ö') newWord += 'Ö';
                else if (presentChar == 'ü') newWord += 'Ü';
                else
                {
                    newWord += presentChar;
                }
            }
            //if a lowercase letter is needed
            else  
            {
                //convert uppercase to lowercase
                if (((int) presentChar >= 65)&&((int) presentChar <= 90))
                {
                    newWord += (char) ((int)presentChar + 32);
                }
                else if (presentChar == 'Ä') newWord += 'ä';
                else if (presentChar == 'Ö') newWord += 'ö';
                else if (presentChar == 'Ü') newWord += 'ü';
                else
                {
                    newWord += presentChar;
                }
            }
        }
        return newWord;
    }
    
    /**
     * used to display a message (debugging or progress info) in System.out or in the log window of the GUI
     * @param output - the logging message to be displayed in System.out or in the log window of the GUI
     */
    public void log(String output)
    {
        //if a Drop was specified, let it handle the output in the GUI
        if (logDrop != null)
        {
            logDrop.put(output + "\n");
        }
        //if no Drop was specified by the constructor, simply print out to System.out
        else
        {
            System.out.println(output);
        }
    }
}