import java.io.*;
import java.util.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import javax.xml.parsers.*;

/**
 * This SAX parser reads through a whole XML document once, reading all the specified "orth" elements 
 * and putting out their content seperated by whitespaces (or by linebreaks) to System.out.
 * Mainly used to extract samples of masked or unmasked content from corpora files.
 * @author Johannes Dellert <jdellert@sfs.uni-tuebingen.de>
 * @version 0.1, November 2007
 */

public class TextExtractor extends DefaultHandler 
{
    //necessary to know where we are during sequential parsing
    boolean isOrth;
    int numberOfTokens;
    int startTokenNumber;
    int endTokenNumber;

    //this string specifies where in the input document the decisive information is to be found
    String orthPath = "../orth";

    //this tracker is used to record curretnt document position during sequential processing
    XPathTracker tracker;

    //used to store whether with line break
    boolean lineBreak;

    //constructor for the default handler
    public TextExtractor(int startNumber, int endNumber, boolean ifLineBreak, String theOrthPath)
    {
        super();
        isOrth = false;
        startTokenNumber = startNumber;
        endTokenNumber = endNumber;
        numberOfTokens = 0;
        lineBreak = ifLineBreak;
        orthPath = theOrthPath;
        tracker = new XPathTracker();
    }

    //print out header
    public void startDocument()
    {
        System.out.print("File contains the following text");
        if (startTokenNumber != -1) 
        {
            System.out.print(" from token " + startTokenNumber);
        }
        if (endTokenNumber != -1) 
        {
            System.out.print(" to token " + endTokenNumber);
        }
        System.out.println(":");
    }

    //count orth tags
    public void startElement(String namespaceURI, String sName, String qName, Attributes attrs)
    {
        tracker.goToElement(qName);
        if (tracker.atPath(orthPath)) 
        {
            isOrth = true;
            numberOfTokens++;
        }
        if (attrs.getLength() >= 1)
        {
            for (int i = 0; i < attrs.getLength(); i++)
            {
                tracker.goToAttribute(attrs.getQName(i));
                if (tracker.atPath(orthPath)) 
                {
                    System.out.print(new String(attrs.getValue(i)));
                    numberOfTokens++;
                    //separate tokens by linebreaks ...
                    if (lineBreak) System.out.print("\n");
                    // ... or blank spaces
                    else System.out.print(" ");
                }
                tracker.levelUp();
            }
        }
    }

    public void endElement(String namespaceURI, String sName, String qName)
    {
        if (tracker.atPath(orthPath)) 
        {
            isOrth = false;
            //separate tokens by linebreaks ...
            if (lineBreak) System.out.print("\n");
            // ... or blank spaces
            else System.out.print(" ");
        }
        tracker.levelUp();
    }


    public void characters(char[] ch, int start, int length)
    {
        //simply print out content of each orth element ...
        if ((isOrth) && ((numberOfTokens <= endTokenNumber) || (endTokenNumber  == -1)) && (numberOfTokens >= startTokenNumber))
        {
            System.out.print(new String(ch, start, length));
        }
    }

    /**
     * This program serves to output the content at certain positions in an XML document
     * Usage:   TextExtractor (-l) (--userDefPaths) [xmlFile] ((startToken)) (endToken)
     *          Option -l for linebreak after each token
     *          Option --userDefPaths to manually specify extracted elements
     */
    public static void main(String[] args)
    {
        File xmlFile = null;
        int startNumber = -1;
        int endNumber = -1;
        boolean lineBreak = false;

        //where in the input document is the decisive information?
        String orthPath = "../orth";
        //are locations defined by the user?
        boolean userDefPaths = false;

        ///////////////////////////////////////////////////////////////////////////////////
        //parse the arguments
        int argsToBeHandled = args.length;
        int argsProcessed = 0;
        if (args[0].equals("--userDefPaths")) 
        {
            userDefPaths = true;
            argsToBeHandled--;
            argsProcessed++;
        }
        if (args[0+argsProcessed].equals("-l")) 
        {
            lineBreak = true;
            argsToBeHandled--;
            argsProcessed++;
        }
        if (argsToBeHandled == 3)
        {
            xmlFile = new File(args[0+argsProcessed]);
            try
            {
                startNumber = Integer.parseInt(args[1+argsProcessed]);
            }
            catch (Exception e)
            {
                System.out.println("Invalid argument: " + args[1+argsProcessed] + "; integerValue expected");
                System.exit(1);
            }
            try
            {
                endNumber = Integer.parseInt(args[2+argsProcessed]);
            }
            catch (Exception e)
            {
                System.out.println("Invalid argument: " + args[2+argsProcessed] + "; integerValue expected");
                System.exit(1);
            }
        }
        if (argsToBeHandled == 2)
        {
            xmlFile = new File(args[0+argsProcessed]);
            try
            {
                startNumber = Integer.parseInt(args[1+argsProcessed]);
            }
            catch (Exception e)
            {
                System.out.println("Invalid argument: " + args[1+argsProcessed] + "; integerValue expected");
                System.exit(1);
            }
        }
        if (argsToBeHandled == 1)
        {
                xmlFile = new File(args[0+argsProcessed]);
        }
        if (argsToBeHandled == 0 || args.length > 5)
        {
            System.out.println("Usage: TextExtractor (-l) (--userDefPaths) [xmlFile] ((startToken)) (endToken)");
            System.out.println("\t Option -l for linebreak after each token");
            System.out.println("\t Option -userDefPaths to manually specify extracted elements");
            System.exit(1);
        }
        //arguments parsed
        //////////////////////////////////////////////////////////
        //let the user define the position if desired
        if (userDefPaths)
        {
            System.out.print("Please specify the XPath to the elements containing the words: ");
            Scanner tagNameScanner = new Scanner(System.in);
            orthPath = tagNameScanner.next(); 
        }
        //now run the specified SAXParser on the document
        TextExtractor extractor = new TextExtractor(startNumber, endNumber, lineBreak,orthPath);
        boolean loadSuccess = xmlFile.canRead();
        if (loadSuccess)
        {
            try
            {
                SAXParserFactory factory = SAXParserFactory.newInstance();
                SAXParser saxParser = factory.newSAXParser();
                saxParser.parse(xmlFile, extractor);
            }
            catch (Throwable t)
            {
                t.printStackTrace();
            }
        }
        else
        {
            System.out.println("File does not exist!");
        }
    }
}